@@ -11,12 +11,8 @@ from .AviStreams import AviMediaReader
 # TODO multi-stream write
 # TODO Buffered reverse video read
 
-__all__ = [
-    "SimpleVideoReader",
-    "SimpleVideoWriter",
-    "SimpleAudioReader",
-    "SimpleAudioWriter",
-    "SimpleVideoFilter",
-    "SimpleAudioFilter",
-    "AviMediaReader",
-]
+# fmt: off
+__all__ = ["SimpleVideoReader", "SimpleVideoWriter", "SimpleAudioReader",
+    "SimpleAudioWriter", "SimpleVideoFilter", "SimpleAudioFilter",
+    "AviMediaReader"]
+# fmt: on
